#!/usr/bin/perl -w 

#
#   Program will use frameshift datafile and input fasta file and will
# try to correct any frameshift that it encounters and that will produce
# a protein larger than the original
#
# WORKFLOW
# loop thru each line in data file
# grab line data
# grab fasta data from fasta DB
# correct
# write out to new fasta file
#
# REQUIRES
#   fasta_tool
#
# CHANGES
# 10/21/13, rmf
#   updated usage instructions
#   added proper command line switches and switch processing
#
# 11/22/13, rmf  
#   fixed so that fasta_tool does not need to be in PATH, but is in bin/util directory
#


use strict;
use Bio::DB::Fasta;
use Bio::SeqIO;
use FindBin;

my $usage 							= qq (
correct_fs_from_fs_data.pl -i infile -d datafile [-o outfile] [-h]

Will read in frameshift detection data and individual fasta entries. Corrections will only be made
if translated correction is longer than translated original. Final output file is both corrected 
transcripts and all others from original fasta file.

	-i      input BLASTX report filename
	-d      input frameshift correction data file
	-o      output filename     [default: inputfile + suffix]
			
	-h      this help text

);

# set defaults and parse args...
my ($fasta_file, $data_file,$corrected_file);

# and parse args
parse_args();

my $corrected_IDs_file = "$corrected_file.corrected_IDs.txt";
my @corrected_seqs;

    #print "$fasta_file, $data_file, $corrected_file, $corrected_IDs_file\n";
    #exit;

# index and make available the FASTA data
my $fasta_db = Bio::DB::Fasta->new($fasta_file);
# open output for corrected file
my $corrected_fasta_io = Bio::SeqIO->new('-format'  => 'fasta',
                                         '-file'    => ">$corrected_file");
# and open input data file...
open (my $DATAFH, "<$data_file") || die "Error opening file $data_file for reading: $!\n";

# now loop thru fs data and correct...
my $counter = 0;
while (my $line = <$DATAFH>) {
    #
    $counter++;
    next if $line =~ m/^(#|\s)/;
    chomp $line;
    # grab data and ensure that we're only looking at the 1st & 2nd HSP
    my @data = split /\t/, $line;
    next if $data[22] != 2;
    
    my ($query_name, $hsp1_frame, $nhsp_frame, $hsp1_qstart, $hsp1_qend,
            $nhsp_qstart, $nhsp_qend) = @data[0,5,7,8..11];
    
    my $seq_obj = $fasta_db->get_Seq_by_id($query_name);
    
    # now adjust seq and write back to seq object
    my $new_seq = adjust_sequence($seq_obj->seq(), $hsp1_frame, $nhsp_frame,
                                    $hsp1_qstart, $hsp1_qend, $nhsp_qstart, $nhsp_qend);
                                    
    my $new_seq_obj = Bio::PrimarySeq->new( -seq => $new_seq, -id => $seq_obj->id(), -description => $seq_obj->description() );
    
    # and write out and remember
    $corrected_fasta_io->write_seq($new_seq_obj);
    push @corrected_seqs, $query_name;

}

close $DATAFH;

# now write out adjusted list
open (my $LISTFH, ">$corrected_IDs_file") || die "Error opening file $corrected_IDs_file for writing: $!\n";
print $LISTFH join ("\n", @corrected_seqs), "\n";
close $LISTFH;

# and filter original file, appending unfiltered seqs to adjusted ones
undef $corrected_fasta_io;      # this closes the corrected fasta file to flush writes to disk
`$FindBin::Bin/../util/fasta_tool --remove $corrected_IDs_file --wrap 60 $fasta_file >> $corrected_file`;

exit;

#
#
# END OF PROGRAM
#

sub parse_args {
	# parse arguments
	while (scalar(@ARGV)) {
		my $arg						= shift @ARGV ;
		if 	  ($arg eq '-h') 		{ die $usage;                       } 
		elsif ($arg eq '-i') 		{ $fasta_file 		= shift @ARGV;   }
		elsif ($arg eq '-d') 		{ $data_file 	    = shift @ARGV;   }
		elsif ($arg eq '-o') 		{ $corrected_file   = shift @ARGV;   }
		else 						{ die "unknown argument '$arg'";    }
	}
	
	# -i input file
	if (!defined $fasta_file) {
		print "You must supply an input filename using '-i' argument!\n";
		die $usage;
	};
	
	# -o output file
	# if not supplied, we're using the default one already set
    $corrected_file = $corrected_file ||  "$fasta_file.corrected";
}





sub adjust_sequence {
    #
    my ($seq, $hsp1_frame, $nhsp_frame,
        $hsp1_qstart, $hsp1_qend, $nhsp_qstart, $nhsp_qend) = @_;
    #
    # four possible configs...
    if ($nhsp_qstart > $hsp1_qstart and $nhsp_qstart <= $hsp1_qend) {
        # hsp1  ------------------
        # nhsp                  -----------
        # fix at nhsp_qstart 
        # frame is still original frame

        # check how to fix seq at 2nd hsp
        if (($hsp1_frame + 1) % 3 == $nhsp_frame) {
            # remove one base at codon hsp1_qend
            substr($seq, $hsp1_qend-3, 1) = '';
        }
        elsif (($hsp1_frame - 1) % 3 == $nhsp_frame) {
            # add one base at codon hsp1_qend
            # need to be one less to insert in front of (vs after) indicated position
            substr($seq, 0, $hsp1_qend-3-1) .= 'N';
        }
        else {
            # throw error
            die "Error in determining frame adjustment between two HSPs!\n";
        }

    }
    elsif ($nhsp_qstart > $hsp1_qstart and $nhsp_qstart > $hsp1_qend) {
        #   ------------------
        #                         -----------
        # concatenate
        # frame is still ok
        $seq = substr($seq, 0, $hsp1_qend) . substr($seq, $nhsp_qstart - 1);
        
    }
    elsif ($nhsp_qend >= $hsp1_qstart and $nhsp_qend < $hsp1_qend) {
        # hsp1            ------------------
        # nhsp  ------------
        # fix at codon of nhsp_qend
        # verify frame is still intact
        # check how to fix seq at 2nd hsp
        if (($hsp1_frame + 1) % 3 == $nhsp_frame) {
            # one base had been removed at junction; 
            # need to add one at codon nhsp_qend
            # and remove one at front to preserve frame of hsp1
            substr($seq, $nhsp_qend - 3 - 1, 1) .= 'N';
            substr($seq, 0, 1) = '';
            
        }
        elsif (($hsp1_frame - 1) % 3 == $nhsp_frame) {
            # one base had been added; 
            # need to remove one base at codong nhsp_qend
            # and add one base at front to preserve frame of hsp1
            substr($seq, $nhsp_qend - 3, 1) = '';
            substr($seq, 0, 1) .= 'N';
        }
        else {
            # throw error
            die "Error in determining frame adjustment between two HSPs!\n";
        }

    }
    elsif ($nhsp_qend < $hsp1_qstart and $nhsp_qend < $hsp1_qend) {
        # hsp1                 ------------------
        # nhsp  ------------
        # concatenate, then adjust frame like previous conditional
        $seq = substr($seq, 0, $nhsp_qend) . substr($seq, $hsp1_qstart - 1);
        
        if (($hsp1_frame + 1) % 3 == $nhsp_frame) {
            # and remove one at front to preserve frame of hsp1
            substr($seq, 0, 1) = '';
        }
        elsif (($hsp1_frame - 1) % 3 == $nhsp_frame) {
            # and add one base at front to preserve frame
            substr($seq, 0, 1) .= 'N';
        }
        else {
            # throw error
            die "Error in determining frame adjustment between two HSPs!\n";
        }

    }
    else {
        die "Error in coordinate overlap determination: "
                . join ("\t", $hsp1_qstart,$hsp1_qend,$nhsp_qstart,$nhsp_qend) . "\n";
    }
    return $seq;
}
