#!/usr/bin/perl -w

#
# This script will parse a multiBLAST report (one file containing multiple
# BLAST reports) and output the information into a tab-delimited file.
#
#
# Output order is:
# 	0	report # in file
# 	1	query name
# 	2	result->query length
# 	3	hit->accession #
# 	4	hit->significance
# 	5	hsp->% ID
# 	6	hsp->numerator # ID
# 	7	hsp->hsp length
# 	8	hit->description
# 	9	hit->rank in report
# added...
#	10  bit score
#	11  gaps
#	12  mismatches
#	13  query start
#	14  query end
#	15  hit start
#	16  hit end
#	17	query strand
#	18	query frame
#   19  subject length
#   20  hsp rank
#
# *** MORE DESCRIPTION NEEDED HERE! ESPECIALLY PARAMETERS AND (ON-SCREEN) OUTPUT

# CHANGELOG
#
# 10/18/13, initial paredown of parent file
#   removed debugging code and option
#   updated usage statement
#   removed requirement for suffix
#   fixed -l and -n parameters code
# 10/24/13
#   fixed warning issued by use of $reportSuffix when not defined
# 4/4/13
#   BLAST parser does not seem to like the new BLAST executables and their
#    output format. Need to grab sequence name from description line. Result->query_name is broken

use strict;
use Bio::SearchIO;
use Cwd;
use English	qw ( -no_match_vars );

$OUTPUT_AUTOFLUSH 				= 1;

my $usage 							= qq (
	parse_multiblast_file.pl -i infile

	Will read in standard (text) NCBI BLAST alignment report and output data in
	a tab-delimited file. Input filename is required

		-i		input file
		
		-o		output file suffix 		[default: .out]
		-l		generate .log file		[default: .log]
				set to 0 to skip
		-m		max hits to process		[default: 20]
		-n		generate .nohits file	[default: .nohits]
				set to 0 to skip
		-x      BLAST file is in XML format
				
		-h      this help text

);

# undef defaults
my ($inFile, $reportSuffix);
# defined defaults
my $currentDir				= cwd();
my $outFile 				= ".out";
my $logFile					= ".log";
my $no_hits_file			= ".nohits";
my $max_hits				= 20;
my $query_count             = 0;							# number of files processed
my $blastformat             = 'blast';

# and parse args
parse_args();

Log("Starting parse_multiblast_file...\n");
Log("\$inFile\t\t$inFile\n");
Log("\$outFile\t\t$outFile\n");
Log("\$logFile\t\t$logFile\n");
Log("\$no_hits_file\t$no_hits_file\n");
# CHANGE 10/24/13, rmf
Log("\$reportSuffix\t$reportSuffix\n") if defined $reportSuffix;

open (GOOD, ">$outFile") || die "Cannot open output file $outFile: $!\n";
my $NOHITS;
if ($no_hits_file) {
	open ($NOHITS, ">$no_hits_file") || die "Cannot open file $no_hits_file for writing: $!\n";
}

wantedReport($inFile);

close GOOD;
close $NOHITS if $no_hits_file;

print "Done!\n";

exit;

#
#
# end of program
#
#

sub wantedReport {

	my $currentFile 					= shift;
	
	Log("Parsing file $currentFile\n");
	my $in = new Bio::SearchIO(	'-format'	=> $blastformat,
								'-file'		=> $currentFile);
								
	my $counter							= 0;				# count w/i the file
	while (my $result = $in->next_result) {
	
		print "." if ++$query_count % 50 == 0;
		++$counter;

        ### CHANGE: RMF, 4/3/14
        #
		my ($query_name) = $result->query_description =~ m/^(\S+)/;
		#

		Log("Parsing query #$counter: $query_name\n");
		
		# no hits?
		if ($result->hits == 0) {
		
			# Log this query & output to NoHits file
			print $NOHITS "$query_name\n" if $no_hits_file;
			Log("NO BLAST HITS FOR $query_name\n");			# keep NO HITS comment on same line
			
		} else {
			
			my $loop_max = ($result->hits <= $max_hits ? $result->hits : $max_hits);
			my $throw_away 				= 0;				# flag to throw the rest of the hits away
			my $hit_counter				= 0;				# our hit count w/i this result
			
			for (1 .. $loop_max) {
				
				my $hit 				= $result->next_hit;
				++$hit_counter;
				my $fh = *GOOD;								# default to *GOOD list
				Print_hit_data($fh, $counter, $result, $hit, $hit_counter);

			} # for $loop (& $hit)

		}	# ($result->hits == 0) check for no hits on this result

		# testing
		#exit if $counter == 3;

	}	# my $result	
}

sub Print_hit_data {
	# prints the query, hit, and hsp data out to the appropriate output file
	# 
	# IN:	$fh			scalar, typeglob of output file
	#		$counter	scalar, report # in file
	#		$result		scalar, (ref/object), result object in file
	#		$hit		scalar (reference/object), hit object to traverse
	# OUT:	nothing
	
	my ($fh, $counter, $result, $hit, $hit_rank) = @_;
	
	# output order:
	# iteration, query_acc#, query_length, hit_acc#, E, %ID, Numerator_ID, Denom_ID, hit_desc, rank

	# hit/hsp
	while (my $hsp 	= $hit->next_hsp()) {
  
        print $fh $counter,"\t";
        
        # query
        my $query_name;
        if ($result->query_name =~ /gnl\|.+\|(.+)/i) {
            $query_name 					= $1;			
        } else {
            ### CHANGE: RMF, 4/3/14
            #$query_name 					= $result->query_name;
             ($query_name) = $result->query_description =~ m/^(\S+)/;
            # END CHANGE
        }
        print {$fh} $query_name,"\t";
        print {$fh} $result->query_length(),"\t";
    
        # hit
        print {$fh} $hit->accession(),"\t";
        print {$fh} $hit->significance(),"\t";
    
        #
        # CHANGE
        # RMF, 9/6/07
        # hsp->%_ID was removed; substituted by 100 * frac_identical
        #print $fh $hsp->percent_identity(),"\t";
        print {$fh} 100 * $hsp->frac_identical(),	"\t";
        print {$fh} $hsp->num_identical(),			"\t";
        print {$fh} $hsp->hsp_length(),				"\t";
        print {$fh} $hit->description(),			"\t";
        #print {$fh} "$hit_rank", 					"\n";		# added 2/6/04, RMF	
        # changed 2/9/11, rmf
        # add output for bit score, # gaps, # mismatches, and alignment coords
        print {$fh} "$hit_rank", 					"\t";
        print {$fh} $hsp->bits(), 					"\t";
        print {$fh} $hsp->gaps(), 					"\t";
        print {$fh} $hsp->length('total')
                        - ($hsp->gaps() + $hsp->num_identical()),	"\t";	# number of mismatches
        print {$fh} $hsp->start('query'), 			"\t";
        print {$fh} $hsp->end('query'), 			"\t";
        print {$fh} $hsp->start('hit'), 			"\t";
        print {$fh} $hsp->end('hit'),	 			"\t";
        print {$fh} $hsp->strand('query'), 			"\t";
        print {$fh} $hsp->query->frame(),	 		"\t";
        print {$fh} $hit->length(),     	 		"\t";
    
        print {$fh} $hsp->rank(),                   "\n";
	}
}


sub Log {
	my $timestamp 					= TimeStamp();
	if ($logFile) {
		# write to a real log file
		open (my $L, ">>$logFile");
		print {$L} "$timestamp: " . $_[0];
		close $L;
	} else {	
		# write to STDERR
		print STDERR "$timestamp: " . $_[0];
	}
}


sub TimeStamp {
  my @mytime 						= localtime(time);
  my $yr 							= 1900 + $mytime[5];
  my $mo							= ZeroPad(1 + $mytime[4]);
  my $da 							= ZeroPad($mytime[3]);
  my $hr 							= ZeroPad($mytime[2]);
  my $mi 							= ZeroPad($mytime[1]);
  my $se 							= ZeroPad($mytime[0]);
  return "$yr$mo$da $hr:$mi:$se";
}


sub ZeroPad {
  return ($_[0]<10?"0$_[0]":$_[0]);
}


sub TimeWord {
  my @mytime 						= localtime(time);
  my $yr 							= 1900 + $mytime[5];
  my $mo 							= ("Jan","Feb","Mar","Apr","May","Jun","Jul","Aug","Sep","Oct","Nov","Dec")[$mytime[4]];
  my $wday 							= ("Sun","Mon","Tue","Wed","Thu","Fri","Sat")[$mytime[6]];
  my $da 							= $mytime[3];
  my $hr 							= $mytime[2];
  my $mi 							= $mytime[1];
  my $se 							= $mytime[0];
  return "$wday, $mo $da, $yr at $hr:$mi:$se";
}


sub parse_args {
	# parse arguments
	while (scalar(@ARGV)) {
		my $arg						= shift @ARGV ;
		if 	  ($arg eq '-h') 		{ die $usage;                    } 
		elsif ($arg eq '-i') 		{ $inFile 		= shift @ARGV;   }
		elsif ($arg eq '-o') 		{ $outFile	 	= shift @ARGV;   }
		elsif ($arg eq '-l') 		{ $logFile  	= shift @ARGV;   }
		elsif ($arg eq '-m') 		{ $max_hits 	= shift @ARGV;   }
		elsif ($arg eq '-n')		{ $no_hits_file = shift @ARGV;   }
		elsif ($arg eq '-x')        { $blastformat  = 'blastxml'     }
		else 						{ die "unknown argument '$arg'"; }
	}
	
	# -i input file
	if (!defined $inFile) {
		print "You must supply an input filename using '-i' argument!\n";
		die $usage;
	};
	
	$outFile 					= $inFile . $outFile;
	#
	$logFile        ? $logFile      = $inFile . $logFile        : undef $logFile;
	$no_hits_file   ? $no_hits_file = $inFile . $no_hits_file   : undef $no_hits_file;

}
